'''
General metric utilities and constants.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
from .format import BoundaryFormat


METRIC_DEFAULTS = {
    'boundary_format': BoundaryFormat.mass,
    'permuted': False,
    'one_minus': False,
    'return_parts': False
}



'''
Tests similarity functions.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
import unittest
from decimal import Decimal
from . import boundary_confusion_matrix, boundary_statistics
from ..format import BoundaryFormat
from ..data.samples import HEARST_1997_STARGAZER, HYPOTHESIS_STARGAZER
from ..ml import precision, recall, fmeasure


class TestSimilarity(unittest.TestCase):

    '''
    Test similarity helper functions.
    '''

    def test_boundary_confusion_matrix(self):
        '''
        Test confusion matrix.
        '''
        cm = boundary_confusion_matrix(
            [set([]), set([2]), set([]), set([]), set([1]), set([1]),
             set([1]), set([1])],
            [set([1]), set([1]), set([]), set([1]), set([]), set([1]),
             set([]), set([])],
            boundary_format=BoundaryFormat.sets)
        self.assertEqual(cm[None][1], 2)
        self.assertEqual(cm[1][None], 1)
        self.assertEqual(cm[None][2], 0)
        self.assertEqual(cm[2][None], 0)
        self.assertEqual(cm[2][1], 1)
        self.assertEqual(cm[1][2], 0)
        self.assertEqual(cm[1][1], Decimal('1.5'))
        self.assertEqual(cm[2][2], 0)

    def test_boundary_statistics(self):
        '''
        Test boundary statistics.
        '''
        value = boundary_statistics([2, 3, 6], [5, 6])
        self.assertEqual(
            {'matches': [1],
             'boundaries_all': 3,
             'pbs': 10,
             'transpositions': [],
             'full_misses': [1],
             'additions': [(1, 'a')],
             'count_edits': Decimal('1'),
             'substitutions': []}, value)

    def test_bed_confusion_matrix(self):
        '''
        Test BED-based confusion matrix upon two segmentations.
        '''
        hypothesis = (5,5,5,5,1)
        reference = HEARST_1997_STARGAZER['stargazer']['2']
        value = boundary_confusion_matrix(hypothesis, reference)
        self.assertAlmostEquals(float(precision(value)), 0.23076923)
        self.assertAlmostEquals(float(recall(value)), 0.23076923)
        self.assertAlmostEquals(float(fmeasure(value)), 0.375)

    def test_bed_confusion_matrix_datasets(self):
        '''
        Test BED-based confusion matrix upon a dataset.
        '''
        hypothesis = HYPOTHESIS_STARGAZER
        reference = HEARST_1997_STARGAZER
        value = boundary_confusion_matrix(hypothesis, reference)
        hyp_p = precision(value)
        hyp_r = recall(value)
        hyp_f = fmeasure(value)
        # Precision
        self.assertAlmostEquals(float(hyp_p['stargazer,h1,1']), 0.57142857)
        self.assertAlmostEquals(float(hyp_p['stargazer,h2,1']), 0.41176470)
        self.assertAlmostEquals(float(hyp_p['stargazer,h1,2']), 0.42857142)
        self.assertAlmostEquals(float(hyp_p['stargazer,h2,2']), 0.33333333)
        # Recall
        self.assertAlmostEquals(float(hyp_r['stargazer,h1,1']), 0.57142857)
        self.assertAlmostEquals(float(hyp_r['stargazer,h2,1']), 0.41176470)
        self.assertAlmostEquals(float(hyp_r['stargazer,h1,2']), 0.42857142)
        self.assertAlmostEquals(float(hyp_r['stargazer,h2,2']), 0.33333333)
        # FMeasure
        self.assertAlmostEquals(float(hyp_f['stargazer,h1,1']), 0.72727272)
        self.assertAlmostEquals(float(hyp_f['stargazer,h2,1']), 0.58333333)
        self.assertAlmostEquals(float(hyp_f['stargazer,h1,2']), 0.6)
        self.assertAlmostEquals(float(hyp_f['stargazer,h2,2']), 0.5)



'''
Segmentation Similarity (S) package.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
from __future__ import division
from . import __boundary_statistics__, SIMILARITY_METRIC_DEFAULTS
from ..util import __fnc_metric__
from decimal import Decimal


def __segmentation_similarity__(*args, **kwargs):
    '''
    Segmentation Similarity (S).
    '''

    metric_kwargs = dict(kwargs)
    del metric_kwargs['return_parts']
    del metric_kwargs['one_minus']
    # Arguments
    boundary_types = kwargs['boundary_types']
    return_parts = kwargs['return_parts']
    one_minus = kwargs['one_minus']
    # Compute
    statistics = __boundary_statistics__(*args, **metric_kwargs)
    # Process
    pbs = statistics['pbs'] * len(boundary_types)
    # Fraction
    denominator = pbs
    numerator = pbs - statistics['count_edits']
    if return_parts:
        return numerator, denominator
    else:
        value = numerator / denominator if denominator > 0 else 1
        if one_minus:
            return Decimal('1') - value
        else:
            return value


def segmentation_similarity(*args, **kwargs):
    '''
    Segmentation Similarity (S).
    '''
    return __fnc_metric__(__segmentation_similarity__, args, kwargs,
                          SIMILARITY_METRIC_DEFAULTS)



'''
Abstract computation utilities.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
from .util.math import mean, std, var, stderr
from itertools import combinations


def compute_pairwise_values(fnc_metric, dataset_a, dataset_b=None, **kwargs):
    '''
    Calculate mean pairwise segmentation metric pairs for functions that take
    pairs of segmentations.

    :param dataset: Segmentation mass dataset (including multiple \
                           codings).
    :param fnc_metric:     Metric function to call on segmentation mass pairs.
    :param permuted:       Permute coder combinations if true.
    :type dataset: dict
    :type fnc_metric:     func
    :type permuted:       bool
    '''

    pairs = dict()
    fnc_kwargs = dict(kwargs)
    # Obtain parameters
    permuted = fnc_kwargs['permuted']
    return_parts = fnc_kwargs['return_parts'] \
        if 'return_parts' in fnc_kwargs else False
    del fnc_kwargs['permuted']

    # Define fnc per group
    def __per_group__(prefix, inner_dataset_m, inner_dataset_n, has_two_datasets):
        '''
        Recurse through a dict to find levels where a metric can be calculated.

        :param inner_dataset: Segmentation mass dataset (including \
                                     multiple codings).
        :type inner_dataset: dict
        '''

        labels = set(inner_dataset_m.keys())
        if inner_dataset_n is not None:
            labels.update(inner_dataset_n.keys())
        for label in labels:
            # Get coders from both datasets
            coder_masses_m = inner_dataset_m[
                label] if label in inner_dataset_m else None
            coder_masses_n = inner_dataset_n[
                label] if inner_dataset_n is not None and label in inner_dataset_n else None
            # Skip this label if it is not contained within both datasets
            if has_two_datasets and (coder_masses_m is None or coder_masses_n is None):
                continue
            # Check to see if there is further nesting
            label_pairs = dict()
            # If is a group
            coder_pairs = None
            if has_two_datasets:
                coders_m = coder_masses_m.keys()
                coders_n = coder_masses_n.keys()
                coder_pairs = [(m, n) for m in coders_m for n in coders_n]
            else:
                coders = coder_masses_m.keys()
                coder_pairs = combinations(coders, 2)
                # We use the same data for both coders
                coder_masses_n = coder_masses_m
            for m, n in coder_pairs:
                segs_m = coder_masses_m[m]
                segs_n = coder_masses_n[n]
                entry_parts = list(prefix)
                entry_parts.extend([label, str(m), str(n)])
                entry = ','.join(entry_parts)
                if return_parts:
                    label_pairs[entry] = fnc_metric(segs_m, segs_n,
                                                    **fnc_kwargs)
                else:
                    label_pairs[entry] = fnc_metric(segs_m, segs_n,
                                                    **fnc_kwargs)
                # Handle permutation
                if permuted and not has_two_datasets:
                    entry_parts = list(prefix)
                    entry_parts.extend([label, str(n), str(m)])
                    entry = ','.join(entry_parts)
                    if return_parts:
                        label_pairs[entry] = \
                            fnc_metric(segs_n, segs_m, **fnc_kwargs)
                    else:
                        label_pairs[entry] = fnc_metric(segs_n, segs_m,
                                                        **fnc_kwargs)
            # Add all
            for entry, pair in label_pairs.items():
                pairs[entry] = pair
            # Erase
            label_pairs = dict()
    # Parse
    has_two_datasets = dataset_b is not None
    __per_group__(tuple(), dataset_a, dataset_b, has_two_datasets)
    # Return mean, std dev, and variance
    return pairs


def summarize(pairs):
    '''
    Takes a list of values and returns the mean, standard deviation, variance, standard error, and number of values.

    :param pairs: List of numerical values
    :type pairs: list
    '''
    return mean(pairs.values()), std(pairs.values()), var(pairs.values()), \
        stderr(pairs.values()), len(pairs)



'''
Tests actual agreement.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
import unittest
from decimal import Decimal
from . import actual_agreement_linear
from ..data.samples import (KAZANTSEVA2012_G5, KAZANTSEVA2012_G2,
                            COMPLETE_AGREEMENT, LARGE_DISAGREEMENT)


class TestAgreement(unittest.TestCase):

    '''
    Test actual agreement.
    '''

    kwargs = {'return_parts': False}

    def test_agreement_g5(self):
        '''
        Test agreement upon Group 5 of Kazantseva (2012) data.
        '''
        agreement = actual_agreement_linear(KAZANTSEVA2012_G5)
        self.assertTrue(agreement > 0)
        self.assertEqual(agreement,
                         Decimal('0.2564575645756457564575645756'))

    def test_agreement_g5_ch1(self):
        '''
        Test agreement upon Group 5, Chapter 1, of Kazantseva (2012) data.
        '''
        data = KAZANTSEVA2012_G5['ch1']
        data = {'ch1': data}
        agreement = actual_agreement_linear(data)
        self.assertTrue(agreement > 0)
        self.assertEqual(agreement,
                         Decimal('0.25'))

    def test_agreement_g2(self):
        '''
        Test agreement upon Group 2 of Kazantseva (2012) data.
        '''
        agreement = actual_agreement_linear(KAZANTSEVA2012_G2)
        self.assertTrue(agreement > 0)
        self.assertEqual(agreement,
                         Decimal('0.4201773835920177383592017738'))

    def test_agreement_g2_ch2(self):
        '''
        Test agreement upon Group 2, Chapter 2, of Kazantseva (2012) data.
        '''
        data = KAZANTSEVA2012_G2['ch2']
        data = {'ch2': data}
        agreement = actual_agreement_linear(data)
        self.assertTrue(agreement > 0)
        self.assertEqual(agreement,
                         Decimal('0.5465116279069767441860465116'))

    def test_agreement_complete(self):
        '''
        Test agreement upon a hypothetical dataset containing complete agreement.
        '''
        agreement = actual_agreement_linear(COMPLETE_AGREEMENT)
        self.assertTrue(agreement >= 0)
        self.assertEqual(agreement,
                         Decimal('1'))

    def test_disagreement_large(self):
        '''
        Test agreement upon a hypothetical dataset containing large disagreement.
        '''
        agreement = actual_agreement_linear(LARGE_DISAGREEMENT)
        self.assertTrue(agreement >= 0)
        self.assertEqual(agreement,
                         Decimal('0'))



'''
Test utilities, and tests for said utilities.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
import unittest
from decimal import Decimal
from . import SegmentationMetricError, __fnc_metric__


class TestCase(unittest.TestCase):

    '''
    A test case that supports performing assertAlmostEquals upon lists, tuples,
    or dicts of values.
    '''

    DECIMAL_PLACES = 4

    def assertAlmostEquals(self, first, second, places=DECIMAL_PLACES,
                           msg=None):
        '''
        Automatically converts values to floats.
        '''

        if isinstance(first, dict) and isinstance(second, dict):
            for item in set(list(first.keys()) + list(second.keys())):
                if not msg:
                    msg = '{0} != {1}'.format(first, second)
                if item not in first:
                    raise Exception(
                        '{0} not in {1}; expected {2}'.format(item, first,
                                                              second))
                if item not in second:
                    raise Exception(
                        '{0} not in {1}; expected {2}'.format(item, second,
                                                              first))
                self.assertAlmostEquals(first[item], second[item], places, msg)
        elif (isinstance(first, list) or isinstance(first, tuple)) and \
                (isinstance(second, list) or isinstance(second, tuple)):
            if len(first) != len(second):
                raise Exception(
                    'Size mismatch; {0} != {1}'.format(first, second))
            for item in zip(first, second):
                if not msg:
                    msg = '{0} != {1}'.format(first, second)
                self.assertAlmostEquals(item[0], item[1], places, msg)
        elif not isinstance(first, type(second)):
            if not isinstance(first, float) and not isinstance(first, Decimal) and \
                    not isinstance(second, float) and not isinstance(second, Decimal):
                raise Exception('Type mismatch; {0} != {1}'.format(
                    type(first), type(second)))
        else:
            return unittest.TestCase.assertAlmostEquals(self,
                                                        float(first),
                                                        float(second),
                                                        places=places,
                                                        msg=msg)


class TestTestCase(TestCase):

    '''
    Test the test utilities.
    '''

    def test_almost_equal_values(self):
        '''
        Test a type mistmatch.
        '''
        self.assertRaises(AssertionError, self.assertAlmostEquals,
                          {'a': 1},
                          {'a': 2})

    def test_mismatch(self):
        '''
        Test a type mistmatch.
        '''
        self.assertRaises(Exception, self.assertAlmostEquals, (), {})

    def test_equal_types(self):
        '''
        Test a type mistmatch.
        '''
        self.assertAlmostEquals({}, {})
        self.assertAlmostEquals([], [])
        self.assertAlmostEquals((), ())
        self.assertAlmostEquals((), [])
        self.assertAlmostEquals([], ())
        self.assertAlmostEquals(0.0, 0.0)
        self.assertAlmostEquals(0, 0)
        self.assertAlmostEquals(Decimal('0'), Decimal('0'))
        self.assertAlmostEquals(0.0, 0)
        self.assertAlmostEquals(0, 0.0)
        self.assertAlmostEquals(0, Decimal('0'))
        self.assertAlmostEquals(Decimal('0'), 0)
        self.assertAlmostEquals(0.0, Decimal('0'))
        self.assertAlmostEquals(Decimal('0'), 0.0)

    def test_tuple_first(self):
        self.assertRaises(Exception, self.assertAlmostEquals, (), (1))

    def test_tuple_second(self):
        self.assertRaises(Exception, self.assertAlmostEquals, (1), ())

    def test_list_first(self):
        self.assertRaises(Exception, self.assertAlmostEquals, [], [1])

    def test_list_second(self):
        self.assertRaises(Exception, self.assertAlmostEquals, [1], [])

    def test_dict_first(self):
        self.assertRaises(Exception, self.assertAlmostEquals, {}, {'a': 1})

    def test_dict_second(self):
        self.assertRaises(Exception, self.assertAlmostEquals, {'a': 1}, {})


class UtilTestCase(TestCase):

    '''
    Test the test utilities.
    '''

    def test_fnc_metric_missing_args(self):
        fnc_metric = lambda x: x
        args = []
        kwargs = {}
        kw_defaults = {}
        self.assertRaises(SegmentationMetricError, __fnc_metric__, fnc_metric, args, kwargs, kw_defaults)



'''
JSON output module (for general JSON writing operations).

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
import json
import os
import codecs
from ..util.lang import enum


Field = enum(
    # Property fields
    segmentation_type='segmentation_type',
    # Structural fields
    items='items'
)

SegmentationType = enum(linear='linear')


def __write_json__(filepath, data):
    '''
    Write a JSON file using the given data.
    '''
    # Create a default filename if a dir is specified
    if os.path.isdir(filepath):
        filepath = os.path.join(filepath, 'output.json')
    # Open file
    json_file = codecs.open(filepath, 'w+', 'utf-8')
    try:
        json.dump(data, fp=json_file, sort_keys=True, indent=4)
    finally:
        json_file.close()


def output_linear_mass_json(filepath, dataset):
    '''
    Takes a file path and :class:`Dataset` and serializes it as JSON.

    :param filepath: Path to the mass file containing segment position codings.
    :type filepath: :func:`str`
    '''
    data = {Field.segmentation_type: SegmentationType.linear}
    data[Field.items] = dataset
    data.update(dataset.properties)
    __write_json__(filepath, data)


def input_linear_mass_json(filepath):
    '''
    Reads a file path. Returns segmentation mass codings as a :class:`Dataset`.

    :param filepath: Path to the mass file containing segment position codings.
    :type filepath: :func:`str`
    '''
    from . import Dataset, DataIOError
    dataset = Dataset()
    data = dict()
    # Open file
    json_file = open(filepath, 'rU')
    # Read in file
    try:
        data = json.load(json_file)
    except Exception as exception:
        raise DataIOError(
            'Error occurred processing file: ' + filepath, exception)
    # Check type
    if Field.segmentation_type in data:
        if data[Field.segmentation_type] != SegmentationType.linear:
            raise DataIOError('Segmentation type \'{0}\' expected, but encountered \'{1}\' for file: {2}'
                              .format(SegmentationType.linear, data[Field.segmentation_type], filepath))
    else:
        raise DataIOError(
            'The entry \'segmentation_type\' was expected in JSON for file:' + filepath)
    # Duplicate to store other properties
    dataset.properties = data
    # If separated into multiple items for one coder per file
    if Field.items in data:
        data = data[Field.items]
        # Convert item labels into strings
        for item, coder_masses in data.items():
            dataset[item] = dict()
            for coder, masses in coder_masses.items():
                dataset[item][coder] = tuple(masses)
        # Remove from properties
        del dataset.properties[Field.items]
    else:
        raise DataIOError('Expected an entry \'{0}\' that contained segmentation codings for specific individual texts (i.e., items) in file: {1}'
                          .format(Field.items, filepath))
    return dataset



'''
Machine learning metric package.  This package a variety of traditional machine
learning metrics that have been adapted for use in segmentation.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
from __future__ import division
from decimal import Decimal
from collections import defaultdict
from ..util import SegmentationMetricError
from ..util.math import mean
from ..util.lang import enum


Average = enum('micro', 'macro')


def __value_micro_macro__(fnc, arguments, classification=None,
                          version=Average.micro):

    def __compute__(fnc, classes, arguments, classification, version):
        if classification is None:
            if version is Average.micro:
                # Micro-average
                numerator, denominator = 0, 0
                for classification in classes:
                    arguments['classification'] = classification
                    arguments['return_parts'] = True
                    class_numerator, class_denominator = fnc(**arguments)
                    numerator += class_numerator
                    denominator += class_denominator
                if numerator == 0:
                    return 0
                else:
                    return Decimal(numerator) / denominator
            elif version is Average.macro:
                # Macro-average
                values = list()
                for classification in classes:
                    arguments['classification'] = classification
                    value = fnc(**arguments)
                    values.append(value)
                return mean(values)
            else:
                raise SegmentationMetricError('Unrecognized type of averaging;\
 expected Average.micro or Average.macro')
        else:
            return fnc(**arguments)
    if isinstance(arguments['matrix'], ConfusionMatrix):
        classes = arguments['matrix'].classes()
        return __compute__(fnc, classes, arguments, classification, version)
    else:
        values = dict()
        new_arguments = dict(arguments)
        for label, matrix in arguments['matrix'].items():
            new_arguments['matrix'] = matrix
            classes = matrix.classes()
            values[label] = __compute__(
                fnc, classes, new_arguments, classification, version)
        return values


def __precision__(matrix, classification, return_parts=False):

    predicted = classification
    denominator = 0
    for actual in matrix.classes():
        denominator += matrix[predicted][actual]
    numerator = matrix[classification][classification]
    if return_parts:
        return numerator, denominator
    else:
        if numerator is 0:
            return Decimal(0)
        else:
            return Decimal(numerator) / Decimal(denominator)


def __recall__(matrix, classification, return_parts=False):

    actual = classification
    denominator = 0
    for predicted in matrix.classes():
        denominator += matrix[predicted][actual]
    numerator = matrix[classification][classification]
    if return_parts:
        return numerator, denominator
    else:
        if numerator is 0:
            return Decimal(0)
        else:
            return Decimal(numerator) / Decimal(denominator)


def __fmeasure__(matrix, classification=None, beta=Decimal('1.0'),
                 return_parts=False):
    '''
    Calculate F-measure, also known as F-score.

    .. math::
        \\text{F}_{\\beta}\\text{-measure} = \\frac{(1 + \\beta^2) \\cdot TP}\
        {(1 + \\beta^2) \\cdot TP + \\beta^2 \\cdot FN + FP}

    :param matrix: Confusion matrix
    :param predicted: Precision for this classification label

    :type matrix: :class:`ConfusionMatrix`

    :returns: F-measure.
    :rtype: :class:`decimal.Decimal`
    '''

    class_precision = __precision__(matrix, classification)
    class_recall = __recall__(matrix, classification)
    if not return_parts and (class_precision == 0 or class_recall == 0):
        return 0
    else:
        # Convert to Decimal
        beta = Decimal(str(beta))
        # Calculate terms
        beta2 = beta ** 2
        beta2_1 = Decimal('1.0') + beta2
        numerator = beta2_1 * class_precision * class_recall
        denominator = (beta2 * class_precision) + class_recall
        if return_parts:
            return numerator, denominator
        else:
            return Decimal(numerator) / Decimal(denominator)


def precision(matrix, classification=None, version=Average.micro):
    '''
    Calculate precision.

    :param matrix: Confusion matrix
    :param classification: Classification label to compute this metric for
    :param version: Averaging-method version.

    :type matrix: :class:`ConfusionMatrix`
    :type classification: Any :class:`dict` index
    :type version: :class:`Average`
    '''

    arguments = dict()
    arguments['matrix'] = matrix
    arguments['classification'] = classification
    return __value_micro_macro__(__precision__, arguments, classification, version)


def recall(matrix, classification=None, version=Average.micro):
    '''
    Calculate recall.

    :param matrix: Confusion matrix
    :param classification: Classification label to compute this metric for
    :param version: Averaging-method version.

    :type matrix: :class:`ConfusionMatrix`
    :type classification: Any :class:`dict` index
    :type version: :class:`Average`
    '''

    arguments = dict()
    arguments['matrix'] = matrix
    arguments['classification'] = classification
    return __value_micro_macro__(__recall__, arguments, classification, version)


def fmeasure(matrix, classification=None, beta=Decimal('1.0'),
             version=Average.micro):
    '''
    Calculate FMeasure.

    :param matrix: Confusion matrix
    :param classification: Classification label to compute this metric for
    :param version: Averaging-method version.

    :type matrix: :class:`ConfusionMatrix`
    :type classification: Any :class:`dict` index
    :type version: :class:`Average`
    '''

    arguments = dict()
    arguments['matrix'] = matrix
    arguments['classification'] = classification
    arguments['beta'] = beta
    return __value_micro_macro__(__fmeasure__, arguments, classification, version)


class _InnerConfusionMatrix(defaultdict):

    '''
    Inner dict of the confusion matrix; used to determine when the classes list
    is dirty.
    '''

    def __init__(self, parent):
        self.__parent__ = parent
        defaultdict.__init__(self, int)

    def __setitem__(self, key, value):
        self.__parent__.__dirty_classes__ = True
        defaultdict.__setitem__(self, key, value)


class ConfusionMatrix(dict):

    '''
    A :func:`dict`-like representation of a confusion matrix offering some automation.
    To access/store values, use: ``matrix[predicted][actual]``.
    '''
    __classes__ = set()
    __dirty_classes__ = False

    def __setitem__(self, key, value):
        raise AttributeError('no such method')

    def __getitem__(self, key):
        '''
        Return default dicts and store them so that the following is possible:

        >>> matrix = ConfusionMatrix()
        >>> matrix['a']['b'] += 1
        >>> matrix['a']['b']
        1
        >>> matrix['a']['a'] = 1
        >>> matrix['a']['a']
        1
        >>> matrix['c']['d']
        0

        '''
        value = None
        if key not in self:
            value = _InnerConfusionMatrix(self)
            dict.__setitem__(self, key, value)
        else:
            value = dict.__getitem__(self, key)
        return value

    def classes(self):
        '''
        Retrieve the set of all classes.
        '''
        if self.__dirty_classes__:
            self.__classes__ = set()
            for predicted, values in self.items():
                self.__classes__.add(predicted)
                for actual in values.keys():
                    self.__classes__.add(actual)
            self.__dirty_classes__ = False
        return self.__classes__



'''
Inter-coder agreement statistics.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
from __future__ import division
from ..data import get_coders
from ..similarity import SIMILARITY_METRIC_DEFAULTS
from ..similarity.boundary import boundary_similarity


AGREEMENT_METRIC_DEFAULTS = dict(SIMILARITY_METRIC_DEFAULTS)
AGREEMENT_METRIC_DEFAULTS.update({
    'fnc_compare': boundary_similarity,
    'return_parts': False
})


def __fnc_metric__(fnc_metric, dataset, **kwargs):
    metric_kwargs = dict(AGREEMENT_METRIC_DEFAULTS)
    metric_kwargs.update(kwargs)
    return fnc_metric(dataset, **metric_kwargs)


def __actual_agreement_linear__(dataset, **kwargs):
    '''
    Calculate actual (i.e., observed or :math:`\\text{A}_a`), segmentation
    agreement without accounting for chance, using [ArtsteinPoesio2008]_'s
    formulation as adapted in [FournierInkpen2012]_:

    .. math::
        \\text{A}_a = \\frac{
            \sum_{i \in I} \\text{mass}(i) \cdot \\text{S}(s_{i1},s_{i2})
        }{
            \sum_{i \in I} \\text{mass}(i)
        }

    Or, for more than two coders:

    .. math::
        \\text{A}_a = \\frac{1}{{\\textbf{c} \\choose 2}}
        \sum^{\\textbf{c}-1}_{m=1}
        \sum^{\\textbf{c}}_{n=m+1}
        \\frac{
            \sum_{i \in I} \\text{mass}(i) \cdot \\text{S}(s_{im},s_{in})
        }{
            \sum_{i \in I} \\big( \\text{mass}(i) - 1 \\big)
        }

    Where :math:`\\text{S}(s_{i1},s_{i2})` is defined in
    :func:`segeval.similarity.SegmentationSimilarity.similarity`.

    :param items_masses: Segmentation masses for a collection of items where \
                        each item is multiply coded (all coders code all items).
    :type items_masses: dict

    :returns: Potential boundaries unmoved, all potential boundaries, and the \
              boundaries per coder.
    :rtype: :func:`list`, :func:`list`, :func:`dict`

    An example of the dictionary structure if the ``items_masses``
    parameter is::

            items_masses = {
                'item1' : {
                    'coder1' : [5],
                    'coder2' : [2,3],
                    'coder2' : [1,1,3]
                },
                'item2' : {
                    'coder1' : [8],
                    'coder2' : [4,4],
                    'coder2' : [2,2,4]
                }
            }

    Other real and contrived examples can be found in
    :mod:`segeval.data.Samples`.

    '''

    metric_kwargs = dict(kwargs)
    del metric_kwargs['fnc_compare']
    metric_kwargs['return_parts'] = True
    # Arguments
    fnc_compare = kwargs['fnc_compare']
    return_parts = kwargs['return_parts']
    # Initialize
    all_numerators = list()
    all_denominators = list()
    all_pbs = list()
    coders_boundaries = dict()
    # Obtain the list of coders
    coders = list(get_coders(dataset))
    # For each permutation of coders
    for m in range(0, len(coders) - 1):
        for n in range(m + 1, len(coders)):
            for item in dataset.keys():
                segs_a = dataset[item][coders[m]]
                segs_b = dataset[item][coders[n]]
                # Compute similarity
                numerator, denominator = \
                    fnc_compare(segs_a, segs_b, **metric_kwargs)[0:2]
                # Obtain necessary values
                pbs = sum(segs_a) - 1
                # Add all pbs
                all_numerators.append(numerator)
                all_denominators.append(denominator)
                all_pbs.append(pbs)
                # Create in dicts if not present
                if coders[m] not in coders_boundaries:
                    coders_boundaries[coders[m]] = list()
                if coders[n] not in coders_boundaries:
                    coders_boundaries[coders[n]] = list()
                # Add per-coder values to dicts
                coders_boundaries[coders[m]].append([len(segs_a), pbs])
                coders_boundaries[coders[n]].append([len(segs_b), pbs])
    if return_parts:
        return all_numerators, all_denominators, all_pbs, coders_boundaries
    else:
        return sum(all_numerators) / sum(all_denominators)


def actual_agreement_linear(dataset, **kwargs):

    return __fnc_metric__(__actual_agreement_linear__, dataset, **kwargs)



'''
Tests the data i/o functions and package.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
import os
import unittest
from . import Dataset, load_nested_folders_dict, FILETYPE_JSON, DataIOError
from .samples import (HEARST_1997_STARGAZER, COMPLETE_AGREEMENT,
                      LARGE_DISAGREEMENT)


class TestDataset(unittest.TestCase):

    '''
    Test data i/o functions and package.
    '''

    def test_dataset(self):
        '''
        Test dataset property creation and independence.
        '''
        prop = 'test'
        dataset_a = Dataset()
        dataset_a.properties[prop] = False
        dataset_b = Dataset()
        self.assertFalse(prop in dataset_b.properties)

    def test_add(self):
        '''
        Test ``Dataset.add()``.
        '''
        # Output complete and large disagreement, then merge
        self.assertEqual(set(['an5', 'an6']), LARGE_DISAGREEMENT.coders)
        large_disagreement = LARGE_DISAGREEMENT.copy()
        self.assertEqual(set(['an5', 'an6']), large_disagreement.coders)
        large_disagreement += COMPLETE_AGREEMENT
        self.assertEqual(6, len(large_disagreement.coders))
        self.assertEqual(4, len(large_disagreement))
        self.assertEqual(6, len(large_disagreement['item1']))

    def test_coders(self):
        '''
        Test ``Dataset.add()``.
        '''
        self.assertEqual(set(['an4', 'an1', 'an2', 'an3']), COMPLETE_AGREEMENT.coders)

    def test_dataset_property(self):
        '''
        Test dataset property creation and independence.
        '''
        prop = 'test'
        dataset = Dataset(properties={prop:True})
        self.assertTrue(prop in dataset.properties)
        self.assertTrue(dataset.properties[prop])

    def test_add_coders(self):
        '''
        Test dataset property creation and independence.
        '''
        dataset_a = Dataset({'item1': {'a': [2]}})
        dataset_b = Dataset({'item1': {'b': [2]}})
        dataset_c = Dataset({'item1': {'a': [2], 'b': [2]}})

        self.assertNotEqual(dataset_a, dataset_b)
        self.assertEqual(dataset_a + dataset_b, dataset_c)
        self.assertNotEqual(dataset_a, dataset_b)
        self.assertNotEqual(dataset_a, dataset_c)
        self.assertNotEqual(dataset_b, dataset_c)

    def test_add_codings(self):
        '''
        Test dataset property creation and independence.
        '''
        dataset_a = Dataset({'item1': {'a': [2]}})
        dataset_b = Dataset({'item2': {'a': [2]}})
        dataset_c = Dataset({'item1': {'a': [2]},
                             'item2': {'a': [2]}})

        self.assertNotEqual(dataset_a, dataset_b)
        self.assertEqual(dataset_a + dataset_b, dataset_c)
        self.assertNotEqual(dataset_a, dataset_b)
        self.assertNotEqual(dataset_a, dataset_c)
        self.assertNotEqual(dataset_b, dataset_c)

    def test_add_duplicate_codings(self):
        '''
        Test dataset property creation and independence.
        '''
        dataset_a = Dataset({'item1': {'a': [2]}})
        dataset_b = Dataset({'item1': {'a': [2]}})
        exception = False
        try:
            dataset_a + dataset_b
        except DataIOError:
            exception = True
        self.assertTrue(exception, 'Did not throw DataIOError')


class TestUtils(unittest.TestCase):

    '''
    Test data merge functions.
    '''

    test_data_dir = os.path.split(__file__)[0]

    def test_load_nested_folders_dict(self):
        '''
        Test nested folder dict construction.
        '''
        data_dir = os.path.abspath(os.path.join(self.test_data_dir, '../'))
        dataset = load_nested_folders_dict(data_dir, FILETYPE_JSON)
        self.assertEqual(dataset['data,stargazer'],
                         HEARST_1997_STARGAZER['stargazer'])



'''
Segmentation encoding format converstion utilities.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
from itertools import groupby
from .util.lang import enum


BoundaryFormat = enum(position='position', mass='mass', sets='sets', nltk='nltk')


def convert_positions_to_masses(positions):
    '''
    Convert an ordered sequence of boundary position labels into a
    sequence of segment masses, e.g., ``[1,1,1,1,1,2,2,2,3,3,3,3,3]`` becomes
    ``[5,3,5]``.

    :param segments: Ordered sequence of which segments a unit belongs to.
    :type segments: tuple

    .. deprecated:: 1.0
    '''
    return tuple([len(list(group)) for key, group in groupby(positions)])


def convert_masses_to_positions(masses):
    '''
    Converts a sequence of segment masses into an ordered sequence of section
    labels for each unit, e.g., ``[5,3,5]`` becomes
    ``[1,1,1,1,1,2,2,2,3,3,3,3,3]``.

    :param masses: Segment mass sequence.
    :type masses: tuple
    '''
    sequence = list()
    for i, mass in enumerate(masses):
        sequence.extend([i + 1] * mass)
    return tuple(sequence)


def boundary_string_from_masses(masses):
    '''
    Creates a "boundary string", or sequence of boundary type sets from a list of segment masses, e.g., ``[5,3,5]`` becomes
    ``[(),(),(),(),(1),(),(),(1),(),(),(),()]``.


    :param masses: Segmentation masses.
    :type masses: tuple
    '''
    string = [set() for _ in range(0, sum(masses) - 1)]
    # Iterate over each position
    pos = 0
    for mass in masses:
        cur_pos = pos + mass - 1
        if cur_pos < len(string):
            string[cur_pos].add(1)
        pos += mass
    # Return
    return tuple([frozenset(pb) for pb in string])


def convert_nltk_to_masses(string, boundary_symbol='1'):
    '''
    Convert an `NLTK <http://nltk.org/>`_-formatted segmentation into masses, e.g., ``000001000100000`` becomes
    ``[5,3,5]``.

    For more information, see `nltk.metrics.segmentation <http://nltk.org/api/nltk.metrics.html#module-nltk.metrics.segmentation>`_.

    :param string: NLTK-formatted segmentation.
    :type string: str
    :param boundary_symbol: String that represents a boundary.
    :type boundary_symbol: str
    '''
    masses = [len(segment) + 1 for segment in string.split(boundary_symbol)]
    return tuple(masses)



'''
Tests the machine learning (ML) statistics functions, and ml package.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
import unittest
from decimal import Decimal
from . import (__precision__, precision, __recall__, recall, __fmeasure__,
               fmeasure, ConfusionMatrix as cm, Average)
from ..util import SegmentationMetricError


class TestConfusionMatrix(unittest.TestCase):

    '''
    Confusion matrix tests.
    '''

    def test_matrix_set_add(self):
        '''
        Test matrix.
        '''
        matrix = cm()
        matrix['p']['p'] += 2
        matrix['p']['n'] = 3
        self.assertEqual(matrix['p']['p'], 2)
        self.assertEqual(matrix['p']['n'], 3)
        self.assertEqual(matrix['p']['f'], 0)
        self.assertEqual(matrix['a']['b'], 0)

    def test_setitem(self):
        '''
        Ensure that __setitem__ raises an AttributeError
        '''
        exception = False
        matrix = cm()
        try:
            matrix['a'] = 0
        except AttributeError:
            exception = True
        self.assertTrue(exception, 'AttributeError not raised')

    def test_matrix_classes(self):
        '''
        Test matrix.
        '''
        matrix = cm()
        matrix['p']['p'] += 2
        matrix['p']['n'] = 3
        self.assertEqual(matrix['p']['p'], 2)
        self.assertEqual(matrix['p']['n'], 3)
        self.assertEqual(matrix['p']['f'], 0)
        self.assertEqual(matrix['a']['b'], 0)

        self.assertEqual(matrix.classes(), set(['p', 'n', 'a', 'b', 'f']))


class TestML(unittest.TestCase):

    '''
    Machine-learning metric tests.
    '''

    def test_precision(self):
        '''
        Test precision.
        '''
        matrix = cm()
        matrix['p']['p'] += 1
        matrix['p']['f'] += 1
        self.assertEqual(__precision__(matrix, 'p'), Decimal('0.5'))
        self.assertEqual(__precision__(matrix, 'f'), Decimal('0'))
        self.assertEqual(
            precision(matrix, version=Average.micro), Decimal('0.5'))
        self.assertEqual(
            precision(matrix, version=Average.macro), Decimal('0.25'))
        matrix = cm()
        matrix['p']['p'] += 1
        matrix['p']['f'] += 3
        matrix['f']['p'] += 1
        self.assertEqual(
            precision(matrix, version=Average.micro), Decimal('0.2'))
        self.assertEqual(
            precision(matrix, version=Average.macro), Decimal('0.125'))
        self.assertEqual(__precision__(matrix, 'p'), Decimal('0.25'))
        self.assertEqual(__precision__(matrix, 'f'), Decimal('0'))
        matrix = cm()
        matrix['p']['p'] += 5
        matrix['p']['f'] += 2
        matrix['f']['p'] += 1
        matrix['f']['f'] += 2
        self.assertEqual(
            precision(matrix, version=Average.micro), Decimal('0.7'))
        self.assertAlmostEqual(precision(matrix, version=Average.macro),
                               Decimal('0.69047'), 4)
        self.assertAlmostEqual(__precision__(matrix, 'p'),
                               Decimal('0.71428'), 4)
        self.assertAlmostEqual(__precision__(matrix, 'f'),
                               Decimal('0.66666'), 4)
        matrix = cm()
        matrix['p']['f'] += 2
        self.assertEqual(precision(matrix), Decimal('0'))
        self.assertEqual(__precision__(matrix, 'p'), Decimal('0'))
        self.assertEqual(__precision__(matrix, 'f'), Decimal('0'))
        matrix = cm()
        matrix['p']['p'] += 2
        self.assertEqual(precision(matrix), Decimal('1'))
        self.assertEqual(__precision__(matrix, 'p'), Decimal('1'))
        self.assertEqual(__precision__(matrix, 'f'), Decimal('0'))
        matrix = cm()
        self.assertEqual(precision(matrix), Decimal('0'))
        self.assertEqual(__precision__(matrix, 'p'), Decimal('0'))
        self.assertEqual(__precision__(matrix, 'f'), Decimal('0'))

    def test_recall(self):
        '''
        Test recall.
        '''
        matrix = cm()
        matrix['p']['p'] += 1
        matrix['p']['f'] += 1
        self.assertEqual(__recall__(matrix, 'p'), Decimal('1.0'))
        self.assertEqual(__recall__(matrix, 'f'), Decimal('0'))
        self.assertEqual(recall(matrix, version=Average.micro), Decimal('0.5'))
        self.assertEqual(recall(matrix, version=Average.macro), Decimal('0.5'))
        matrix = cm()
        matrix['p']['p'] += 1
        matrix['p']['f'] += 3
        matrix['f']['p'] += 1
        self.assertEqual(recall(matrix, version=Average.micro), Decimal('0.2'))
        self.assertEqual(
            recall(matrix, version=Average.macro), Decimal('0.25'))
        self.assertEqual(__recall__(matrix, 'p'), Decimal('0.5'))
        self.assertEqual(__recall__(matrix, 'f'), Decimal('0'))
        matrix = cm()
        matrix['p']['p'] += 5
        matrix['p']['f'] += 2
        matrix['f']['p'] += 1
        matrix['f']['f'] += 2
        self.assertEqual(recall(matrix, version=Average.micro), Decimal('0.7'))
        self.assertAlmostEqual(recall(matrix, version=Average.macro),
                               Decimal('0.66666'), 4)
        self.assertAlmostEqual(__recall__(matrix, 'p'),
                               Decimal('0.83333'), 4)
        self.assertAlmostEqual(__recall__(matrix, 'f'),
                               Decimal('0.5'), 4)
        matrix = cm()
        matrix['p']['f'] += 2
        self.assertEqual(recall(matrix), Decimal('0'))
        self.assertEqual(__recall__(matrix, 'p'), Decimal('0'))
        self.assertEqual(__recall__(matrix, 'f'), Decimal('0'))
        matrix = cm()
        matrix['p']['p'] += 2
        self.assertEqual(recall(matrix), Decimal('1'))
        self.assertEqual(__recall__(matrix, 'p'), Decimal('1'))
        self.assertEqual(__recall__(matrix, 'f'), Decimal('0'))
        matrix = cm()
        self.assertEqual(recall(matrix), Decimal('0'))
        self.assertEqual(__recall__(matrix, 'p'), Decimal('0'))
        self.assertEqual(__recall__(matrix, 'f'), Decimal('0'))

    def test_fmeasure(self):
        '''
        Test FMeasure.
        '''
        matrix = cm()
        matrix['p']['p'] += 1
        matrix['p']['f'] += 1
        self.assertAlmostEqual(
            __fmeasure__(matrix, 'p'), Decimal('0.66666'), 4)
        self.assertEqual(__fmeasure__(matrix, 'f'), Decimal('0'))
        self.assertAlmostEqual(fmeasure(matrix, version=Average.micro),
                               Decimal('0.66666'), 4)
        self.assertAlmostEqual(fmeasure(matrix, version=Average.macro),
                               Decimal('0.33333'), 4)
        self.assertAlmostEqual(fmeasure(matrix, classification='p'),
                               Decimal('0.66666'), 4)
        matrix = cm()
        matrix['p']['p'] += 1
        matrix['p']['f'] += 3
        matrix['f']['p'] += 1
        self.assertAlmostEqual(fmeasure(matrix, version=Average.micro),
                               Decimal('0.33333'), 4)
        self.assertAlmostEqual(fmeasure(matrix, version=Average.macro),
                               Decimal('0.16666'), 4)
        self.assertAlmostEqual(__fmeasure__(matrix, 'p'),
                               Decimal('0.33333'), 4)
        self.assertAlmostEqual(fmeasure(matrix, classification='p'),
                               Decimal('0.33333'), 4)
        self.assertEqual(__fmeasure__(matrix, 'f'), Decimal('0'))
        matrix = cm()
        matrix['p']['p'] += 5
        matrix['p']['f'] += 2
        matrix['f']['p'] += 1
        matrix['f']['f'] += 2
        self.assertAlmostEqual(fmeasure(matrix, version=Average.micro),
                               Decimal('0.68421'), 4)
        self.assertAlmostEqual(fmeasure(matrix, version=Average.macro),
                               Decimal('0.67032'), 4)
        self.assertAlmostEqual(__fmeasure__(matrix, 'p'),
                               Decimal('0.76923'), 4)
        self.assertAlmostEqual(__fmeasure__(matrix, 'f'),
                               Decimal('0.57142'), 4)
        matrix = cm()
        matrix['p']['f'] += 2
        self.assertEqual(fmeasure(matrix), Decimal('0'))
        self.assertEqual(__fmeasure__(matrix, 'p'), Decimal('0'))
        self.assertEqual(__fmeasure__(matrix, 'f'), Decimal('0'))
        matrix = cm()
        matrix['p']['p'] += 2
        self.assertEqual(fmeasure(matrix), Decimal('1'))
        self.assertEqual(__fmeasure__(matrix, 'p'), Decimal('1'))
        self.assertEqual(__fmeasure__(matrix, 'f'), Decimal('0'))
        matrix = cm()
        self.assertEqual(fmeasure(matrix), Decimal('0'))
        self.assertEqual(__fmeasure__(matrix, 'p'), Decimal('0'))
        self.assertEqual(__fmeasure__(matrix, 'f'), Decimal('0'))

    def test_exception_on_incorrect_average(self):
        '''
        Test exception on incorrect average.
        '''
        matrix = cm()
        matrix['p']['p'] += 1
        matrix['p']['f'] += 1
        self.assertRaises(
            SegmentationMetricError, fmeasure, matrix, version='incorrect')



'''
Similarity utility functions based upon boundary edit distance.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
from __future__ import division
from .distance.multipleboundary import boundary_edit_distance
from .weight import weight_a, weight_s_scale, weight_t_scale
from ..metric import METRIC_DEFAULTS
from ..ml import ConfusionMatrix as cm
from ..format import (BoundaryFormat, boundary_string_from_masses,
                      convert_positions_to_masses, convert_nltk_to_masses)
from ..util import __fnc_metric__, SegmentationMetricError


SIMILARITY_METRIC_DEFAULTS = dict(METRIC_DEFAULTS)
SIMILARITY_METRIC_DEFAULTS.update({
    'n_t': 2,
    'boundary_types': frozenset([1]),
    'weight': (weight_a, weight_s_scale, weight_t_scale)
})


def __boundary_statistics__(segs_a, segs_b, boundary_types, boundary_format, n_t, weight):
    '''
    Compute boundary similarity applying the weighting functions specified.
    '''

    # Convert from NLTK types
    if boundary_format == BoundaryFormat.nltk:
        segs_a = convert_nltk_to_masses(segs_a)
        segs_b = convert_nltk_to_masses(segs_b)
        boundary_format = BoundaryFormat.mass
    # Check format
    if boundary_format == BoundaryFormat.sets:
        pass  # Correct boundary format
    elif boundary_format == BoundaryFormat.mass:
        segs_a = boundary_string_from_masses(segs_a)
        segs_b = boundary_string_from_masses(segs_b)
    elif boundary_format == BoundaryFormat.position:
        segs_a = convert_positions_to_masses(segs_a)
        segs_b = convert_positions_to_masses(segs_b)
        segs_a = boundary_string_from_masses(segs_a)
        segs_b = boundary_string_from_masses(segs_b)
    else:
        raise SegmentationMetricError('Unsupported boundary format')
    # Check length
    if len(segs_a) != len(segs_b):
        raise SegmentationMetricError(
            'Segmentations differ in length ({0} != {1})'.format(
                len(segs_a), len(segs_b)))
    # Calculate the total pbs
    pbs = len(segs_b) * len(boundary_types)
    # Compute edits
    additions, substitutions, transpositions = \
        boundary_edit_distance(segs_a, segs_b, n_t=n_t)
    # Apply weighting functions
    fnc_weight_a, fnc_weight_s, fnc_weight_t = weight
    count_additions = fnc_weight_a(additions)
    count_substitutions = fnc_weight_s(substitutions,
                                       max(boundary_types),
                                       min(boundary_types))
    count_transpositions = fnc_weight_t(transpositions, n_t)
    count_edits = count_additions + count_substitutions + count_transpositions
    # Compute
    matches = list()
    full_misses = list()
    boundaries_all = 0
    for set_a, set_b in zip(segs_a, segs_b):
        matches.extend(set_a.intersection(set_b))
        full_misses.extend(set_a.symmetric_difference(set_b))
        boundaries_all += len(set_a) + len(set_b)
    return {'count_edits': count_edits, 'additions': additions,
            'substitutions': substitutions, 'transpositions': transpositions,
            'full_misses': full_misses, 'boundaries_all': boundaries_all,
            'matches': matches, 'pbs': pbs}


def __boundary_confusion_matrix__(*args, **kwargs):
    '''
    Create a confusion matrix using boundary edit distance.
    '''

    # Trim kwargs
    metric_kwargs = dict(kwargs)
    del metric_kwargs['return_parts']
    del metric_kwargs['one_minus']
    # Obtain statistics
    statistics = __boundary_statistics__(*args, **metric_kwargs)
    # Get parameters
    n_t = kwargs['n_t']
    weight = kwargs['weight']
    # Initialize
    matrix = cm()
    fnc_weight_t = weight[2]
    # Add matches
    for match in statistics['matches']:
        matrix[match][match] += 1
    # Add weighted near misses
    for transposition in statistics['transpositions']:
        match = transposition[2]
        matrix[match][match] += fnc_weight_t([transposition], n_t)
    # Add confusion errors
    for substitution in statistics['substitutions']:
        hyp, ref = substitution
        matrix[hyp][ref] += 1
    # Add full misses
    for addition in statistics['additions']:
        hyp, ref = None, None
        boundary_type, side = addition
        if side == 'a':
            hyp = None
            ref = boundary_type
        else:  # side == 'b'
            hyp = boundary_type
            ref = None
        assert side == 'a' or side == 'b'
        matrix[hyp][ref] += 1
    return matrix


def boundary_confusion_matrix(*args, **kwargs):

    return __fnc_metric__(__boundary_confusion_matrix__, args, kwargs,
                          SIMILARITY_METRIC_DEFAULTS)


def boundary_statistics(*args, **kwargs):

    default_kwargs = dict(SIMILARITY_METRIC_DEFAULTS)
    del default_kwargs['one_minus']
    del default_kwargs['return_parts']
    return __fnc_metric__(__boundary_statistics__, args, kwargs,
                          default_kwargs)



'''
TSV output module (for general TSV writing operations).

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
import csv
from ..format import convert_positions_to_masses


DEFAULT_DELIMITER = '\t'


def input_linear_mass_tsv(filepath, delimiter=DEFAULT_DELIMITER):
    '''
    Takes a file path.  Returns segmentation mass codings as a :class:`Dataset`.

    :param filepath: path to the mass file containing segment mass codings.
    :param delimiter:    the delimiter used when reading a TSV file (by default,
                         a tab, but it can also be a comma, whitespace, etc.
    :type filepath: str
    :type delimiter: str
    '''

    from . import Dataset, name_from_filepath
    # List version of file
    header = []
    dataset = Dataset()
    item = name_from_filepath(filepath)
    dataset[item] = dict()
    # Open file
    with open(filepath, 'rU') as csv_file:
        # Read in file
        reader = csv.reader(csv_file, delimiter=delimiter)
        for i, row in enumerate(reader):
            # Read annotators from header
            if i is 0:
                for col_name in row[1:]:
                    header.append(col_name)
            # Read data
            else:
                coder = None
                for j, col in enumerate(row):
                    # Skip the first col
                    if j is 0:
                        coder = str(col)
                        dataset[item][coder] = list()
                    else:
                        dataset[item][coder].append(int(col))
                dataset[item][coder] = tuple(dataset[item][coder])

    return dataset


def input_linear_positions_tsv(filepath, delimiter=DEFAULT_DELIMITER):
    '''
    Takes a file path.  Returns segmentation mass codings as a :class:`Dataset`.

    :param filepath: path to the mass file containing segment position
                         codings.
    :param delimiter:    the delimiter used when reading a TSV file (by default,
                         a tab, but it can also be a comma, whitespace, etc.
    :type filepath: str
    :type delimiter: str

    .. deprecated:: 1.0

    .. warning:: This I/O function is for legacy files only and will be removed
        in later versions.
    '''
    dataset = input_linear_mass_tsv(filepath, delimiter)
    # Convert each segment position to masses
    for item, coder_positions in dataset.items():
        for coder, positions in coder_positions.items():
            dataset[item][coder] = convert_positions_to_masses(positions)
    # Return
    return dataset



'''
Tests segmentation format utility functions.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
import unittest
from .format import (convert_positions_to_masses, convert_masses_to_positions,
                     boundary_string_from_masses, convert_nltk_to_masses)


class TestSegeval(unittest.TestCase):

    '''
    Segmentation-format-related function tests.
    '''

    def test_convert_positions_to_masses(self):
        '''
        Test segment position sequence conversion to masses.
        '''
        self.assertEqual((5,3,5),
                         convert_positions_to_masses([1,1,1,1,1,2,2,2,3,3,3,3,3]))

    def test_convert_positions_to_masses_none(self):
        '''
        Test segment position sequence conversion from masses.
        '''
        self.assertEqual((11,),
                         convert_positions_to_masses([1,1,1,1,1,1,1,1,1,1,1]))

    def test_convert_positions_to_masses_all(self):
        '''
        Test segment position sequence conversion from masses.
        '''
        self.assertEqual((1,1,1,1,1,1,1,1,1,1,1),
                         convert_positions_to_masses([1,2,3,4,5,6,7,8,9,10,11]))

    def test_convert_masses_to_positions(self):
        '''
        Test segment position sequence conversion from masses.
        '''
        self.assertEqual((1,1,1,1,1,2,2,2,3,3,3,3,3),
                         convert_masses_to_positions([5,3,5]))

    def test_convert_masses_to_positions_none(self):
        '''
        Test segment position sequence conversion from masses.
        '''
        self.assertEqual(convert_masses_to_positions([11]),
                         (1,1,1,1,1,1,1,1,1,1,1))

    def test_convert_masses_to_positions_all(self):
        '''
        Test segment position sequence conversion from masses.
        '''
        self.assertEqual(convert_masses_to_positions([1,1,1,1,1,1,1,1,1,1,1]),
                         (1,2,3,4,5,6,7,8,9,10,11))

    def test_boundary_string_from_masses_none(self):
        '''
        No boundaries.
        '''
        string = boundary_string_from_masses([3])
        self.assertEqual(string, (set(), set()))

    def test_boundary_string_from_masses_full(self):
        '''
        Few boundaries.
        '''
        string = boundary_string_from_masses([1,1,1,1])
        self.assertEqual(string, (set([1]), set([1]), set([1])))

    def test_boundary_string_from_masses_one(self):
        '''
        Few boundaries.
        '''
        string = boundary_string_from_masses([2,3])
        self.assertEqual(string, (set(), set([1]), set(), set()))

    def test_convert_nltk_to_masses_pk_ab(self):
        '''
        NLTK-style segmentations starting with a boundary.
        '''
        self.assertEqual(convert_nltk_to_masses('100'), (1,3))
        self.assertEqual(convert_nltk_to_masses('010'), (2,2))

    def test_convert_nltk_to_masses_pk_long(self):
        '''
        NLTK-style segmentations starting with a boundary.
        '''
        self.assertEqual(convert_nltk_to_masses('0100100000'), (2, 3, 6))
        self.assertEqual(convert_nltk_to_masses('0101000000'), (2, 2, 7))



'''
Tests window-based segmentation metric utility functions.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
from . import compute_window_size
from ..util import SegmentationMetricError
from ..util.test import TestCase
from ..data.samples import KAZANTSEVA2012_G5


class TestWindow(TestCase):

    '''
    Test window metric fncs.
    '''

    def test_window_size(self):
        reference = [1, 2, 3]
        self.assertEqual(2, compute_window_size(reference))

    def test_window_size_dataset(self):
        self.assertEqual(3, compute_window_size(KAZANTSEVA2012_G5))

    def test_window_size_exception(self):
        reference = 'incorrect type'
        self.assertRaises(SegmentationMetricError, compute_window_size, reference)



'''
Tests some math related utility functions.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
import unittest
from decimal import Decimal
from .math import mean, std, var, stderr


class TestMath(unittest.TestCase):

    '''
    Math utlity function tests.
    '''

    def test_mean(self):
        '''
        Tests population mean.
        '''
        self.assertEqual(5, mean([2, 4, 4, 4, 5, 5, 7, 9]))

    def test_mean_0(self):
        '''
        Tests population mean.
        '''
        self.assertEqual(0, mean([]))

    def test_std(self):
        '''
        Tests population standard deviation.
        '''
        self.assertEqual(2, std([2, 4, 4, 4, 5, 5, 7, 9]))

    def test_var(self):
        '''
        Tests population variance.
        '''
        self.assertEqual(4, var([2, 4, 4, 4, 5, 5, 7, 9]))

    def test_stderr(self):
        '''
        Tests population standard error of the mean.
        '''
        self.assertEqual(Decimal('0.7071067811865475244008443622'),
                         stderr([2, 4, 4, 4, 5, 5, 7, 9]))



'''
Segmentation evaluation metric package. Provides evaluation metrics to
evaluate the performance of both human and automatic text (i.e., discourse)
segmenters.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
import sys
from types import ModuleType


# Package description
__version_number__ = '2.0.8'
__release__ = None
__version__ = '-'.join((__version_number__, __release__)) if __release__ is not None else __version_number__
__project__ = 'SegEval'
__package__ = 'segeval'
__author__ = 'Chris Fournier'
__author_email__ = 'chris.m.fournier@gmail.com'
__copyright__ = '2012-2013, ' + __author__
__description__ = 'A package providing text segmentation evaluation metrics and utilities'


# The import magic shown here was taken (nearly verbatim) from the Werkzeug
# package; see http://werkzeug.pocoo.org/

# import mapping to objects in other modules
all_by_module = {
    'segeval.agreement':        ['actual_agreement_linear'],
    'segeval.agreement.bias':   ['artstein_poesio_bias_linear'],
    'segeval.agreement.kappa':  ['fleiss_kappa_linear'],
    'segeval.agreement.pi':     ['fleiss_pi_linear'],
    'segeval.data':             ['Dataset', 'load_nested_folders_dict'],
    'segeval.data.jsonutils':   ['Field', 'input_linear_mass_json',
                                 'output_linear_mass_json'],
    'segeval.data.tsv':         ['input_linear_mass_tsv'],
    'segeval.data.samples':     ['KAZANTSEVA2012_G5', 'KAZANTSEVA2012_G2',
                                 'COMPLETE_AGREEMENT', 'LARGE_DISAGREEMENT',
                                 'HEARST_1997_STARGAZER', 'HYPOTHESIS_STARGAZER'],
    'segeval.ml':               ['Average', 'precision', 'recall',
                                 'fmeasure', 'ConfusionMatrix'],
    'segeval.similarity':       ['boundary_confusion_matrix',
                                 'boundary_statistics'],
    'segeval.similarity.boundary':
                                ['boundary_similarity'],
    'segeval.similarity.segmentation':
                                ['segmentation_similarity'],
    'segeval.similarity.distance.multipleboundary':
                                ['boundary_edit_distance'],
    'segeval.similarity.weight':
                                ['weight_a',
                                 'weight_s',
                                 'weight_s_scale',
                                 'weight_t',
                                 'weight_t_scale'],
    'segeval.window':           ['compute_window_size'],
    'segeval.window.pk':        ['pk'],
    'segeval.window.windowdiff':['window_diff'],
    'segeval.compute':          ['summarize'],
    'segeval.format':           ['BoundaryFormat',
                                 'boundary_string_from_masses',
                                 'convert_positions_to_masses',
                                 'convert_masses_to_positions',
                                 'convert_nltk_to_masses'],
}


object_origins = {}
for module, items in all_by_module.items():
    for item in items:
        object_origins[item] = module


class module(ModuleType):
    """Automatically import objects from the modules."""

    def __getattr__(self, name):
        if name in object_origins:
            module = __import__(object_origins[name], None, None, [name])
            for extra_name in all_by_module[module.__name__]:
                setattr(self, extra_name, getattr(module, extra_name))
            return getattr(module, name)
        return ModuleType.__getattribute__(self, name)

    def __dir__(self):
        """Just show what we want to show."""
        result = list(new_module.__all__)
        result.extend(
            ('__file__', '__path__', '__doc__', '__all__',
             '__docformat__', '__name__', '__path__', '__package__',
             '__project__', '__version__'))
        return result


# keep a reference to this module so that it's not garbage collected
old_module = sys.modules['segeval']


# setup the new module and patch it into the dict of loaded modules
new_module = sys.modules['segeval'] = module('segeval')
new_module.__dict__.update({
    '__file__':             __file__,
    '__path__':             __path__,
    '__package__':          __package__,
    '__project__':          __project__,
    '__doc__':              __doc__,
    '__version__':          __version__,
    '__version_number__':   __version_number__,
    '__author__':           __author__,
    '__author_email__':     __author_email__,
    '__copyright__':        __copyright__,
    '__all__':              tuple(object_origins),
    '__docformat__':        'restructuredtext en',
    '__description__':      __description__
})



'''
Boundary Similarity (B) package.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
from __future__ import division
from . import __boundary_statistics__, SIMILARITY_METRIC_DEFAULTS
from ..util import __fnc_metric__
from decimal import Decimal


def __boundary_similarity__(*args, **kwargs):

    metric_kwargs = dict(kwargs)
    del metric_kwargs['return_parts']
    del metric_kwargs['one_minus']
    # Arguments
    return_parts = kwargs['return_parts']
    one_minus = kwargs['one_minus']
    # Compute
    statistics = __boundary_statistics__(*args, **metric_kwargs)
    additions = statistics['additions']
    substitutions = statistics['substitutions']
    transpositions = statistics['transpositions']
    count_unweighted = len(additions) + len(substitutions) + len(transpositions)
    # Fraction
    denominator = count_unweighted + len(statistics['matches'])
    numerator = denominator - statistics['count_edits']
    if return_parts:
        return numerator, denominator, additions, substitutions, transpositions
    else:
        value = numerator / denominator if denominator > 0 else 1
        if one_minus:
            return Decimal('1') - value
        else:
            return value


def boundary_similarity(*args, **kwargs):
    '''
    Boundary Similarity (B).
    '''

    return __fnc_metric__(__boundary_similarity__, args, kwargs,
                          SIMILARITY_METRIC_DEFAULTS)



'''
Tests boundary similarity (B).

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
import unittest
from decimal import Decimal
from .boundary import boundary_similarity
from .weight import weight_a, weight_s, weight_t
from ..util import SegmentationMetricError
from ..format import BoundaryFormat
from ..compute import summarize
from ..data.samples import (MULTIPLE_BOUNDARY_TYPES, HEARST_1997_STARGAZER,
                            HYPOTHESIS_STARGAZER)


class TestBoundary(unittest.TestCase):

    '''
    Test segmenation boundary comparison functions.
    '''

    def test_fn(self):
        '''
        Test false negative.
        '''
        value = boundary_similarity([2, 3, 6], [5, 6])
        self.assertEqual(Decimal('0.5'), value)

    def test_fp(self):
        '''
        Test false negative.
        '''
        value = boundary_similarity([2, 3, 6], [2, 3, 3, 3])
        self.assertAlmostEquals(Decimal('0.66666'), value, 4)

    def test_near_miss(self):
        '''
        Test near miss.
        '''
        value = boundary_similarity([2, 3, 6], [2, 2, 7])
        self.assertEqual(Decimal('0.75'), value)

    def test_one_minus(self):
        '''
        Test one minus.
        '''
        value = boundary_similarity([2, 3, 6], [2, 2, 7], one_minus=True)
        self.assertEqual(Decimal('0.25'), value)

    def test_boundary_format_nltk(self):
        '''
        Test the nltk boundary format.
        '''
        value = boundary_similarity('01001000000', '01010000000', boundary_format=BoundaryFormat.nltk)
        self.assertAlmostEqual(Decimal('0.75'), value)

    def test_clustered_fps(self):
        '''
        Test clustered fps.
        '''
        value = boundary_similarity([2, 3, 6], [1, 1, 3, 1, 5])
        self.assertEqual(Decimal('0.5'), value)

    def test_positions(self):
        '''
        Test false negative.
        '''
        a = [1,1,1,1,1,1,1,1,1,1,1,1,1]
        b = [1,1,1,1,2,2,2,2,3,3,3,3,3]
        value = boundary_similarity(a, b, boundary_format=
                                    BoundaryFormat.position)
        self.assertEqual(Decimal('0'), value)

    def test_format_exception(self):
        '''
        Test false negative.
        '''
        a = [1,1,1,1,1,1,1,1,1,1,1,1,1]
        b = [1,1,1,1,2,2,2,2,3,3,3,3,3]
        self.assertRaises(SegmentationMetricError, boundary_similarity, a, b,
                          boundary_format=None)

    def test_arg_exception(self):
        '''
        Test false negative.
        '''
        a = [1,1,1,1,1,1,1,1,1,1,1,1,1]
        b = [1,1,1,1,2,2,2,2,3,3,3,3,3]
        c = 0
        self.assertRaises(SegmentationMetricError, boundary_similarity, a, b,
                          c)

    def test_weight_t(self):
        '''
        Test false negative.
        '''
        value = boundary_similarity([2, 3, 6], [5, 6],
                                    weight=(weight_a, weight_s, weight_t))
        self.assertEqual(0.5, value)

    def test_multiple_boundary_types(self):
        '''
        Test false negative.
        '''
        value = summarize(boundary_similarity(MULTIPLE_BOUNDARY_TYPES))
        self.assertEqual((Decimal('0.375'),
                          Decimal('0.125'),
                          Decimal('0.015625'),
                          Decimal('0.08838834764831844055010554528'),
                          2),
                         value)

    def test_b_datasets(self):
        '''
        Test B upon two datasets.
        '''
        hypothesis = HYPOTHESIS_STARGAZER
        reference = HEARST_1997_STARGAZER
        value = boundary_similarity(hypothesis, reference)

        # Precision
        self.assertAlmostEquals(float(value['stargazer,h1,1']), 0.57142857)
        self.assertAlmostEquals(float(value['stargazer,h2,1']), 0.38888888)
        self.assertAlmostEquals(float(value['stargazer,h1,2']), 0.42857142)
        self.assertAlmostEquals(float(value['stargazer,h2,2']), 0.33333333)



'''
Implementation of the Pk segmentation evaluation metric described in
[BeefermanBerger1999]_.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
from __future__ import division
from decimal import Decimal
from . import __compute_window_size__, WINDOW_METRIC_DEFAULTS
from ..util import __fnc_metric__, SegmentationMetricError
from ..format import (BoundaryFormat, convert_masses_to_positions, convert_nltk_to_masses)


def __pk__(hypothesis, reference, window_size, one_minus, boundary_format,
           return_parts, fnc_round):

    # Convert from NLTK types
    if boundary_format == BoundaryFormat.nltk:
        reference = convert_nltk_to_masses(reference)
        hypothesis = convert_nltk_to_masses(hypothesis)
        boundary_format = BoundaryFormat.mass
    # Convert from masses into positions
    if boundary_format == BoundaryFormat.mass:
        reference = convert_masses_to_positions(reference)
        hypothesis = convert_masses_to_positions(hypothesis)
    elif boundary_format != BoundaryFormat.position:
        raise SegmentationMetricError('Unsupported boundary format')
    # Check for input errors
    if len(reference) is not len(hypothesis):
        raise SegmentationMetricError(
            'Reference and hypothesis segmentations differ in position length ({0} is not {1}).'.format(len(reference), len(hypothesis)))
    # Compute window size to use if unspecified
    if window_size is None:
        window_size = __compute_window_size__(reference, fnc_round,
                                              BoundaryFormat.position)
    # Create a set of pairs of units from each segmentation to go over using a
    # window
    sum_differences = 0
    # Slide window over and sum the number of varying windows
    measurements = 0
    for i in range(0, len(reference) - (window_size)):
        # Create probe windows with k boundaries inside
        window_ref = reference[i:i + window_size + 1]
        window_hyp = hypothesis[i:i + window_size + 1]
        # Probe agreement
        agree_ref = window_ref[0] is window_ref[-1]
        agree_hyp = window_hyp[0] is window_hyp[-1]
        # If the windows agreements agree
        if agree_ref is not agree_hyp:
            sum_differences += 1
        measurements += 1
    # Perform final division
    value = Decimal(sum_differences) / measurements if measurements > 0 else 0
    if return_parts:
        return sum_differences, measurements
    else:
        if one_minus:
            return Decimal('1.0') - value
        else:
            return value


def pk(*args, **kwargs):

    return __fnc_metric__(__pk__, args, kwargs, WINDOW_METRIC_DEFAULTS)



'''
Tests weight functions.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
import unittest
from decimal import Decimal
from . import (weight_s_scale, weight_t_scale)


class TestWeight(unittest.TestCase):

    '''
    Test similarity helper functions.
    '''

    def test_weight_t_scale(self):
        '''
        Test to see that penalties for substitution edits are discounted from
        3 to 1.5.
        '''
        transpositions = [[1,2], [1,3], [1,4]]
        weight = weight_t_scale(transpositions, 4)
        self.assertEqual(weight, Decimal('1.5'))

    def test_weight_s_scale(self):
        '''
        Test to see that penalties for transposition edits are discounted from
        3 to 1.5.
        '''
        substitutions = [[1,2], [1,3], [1,4]]
        weight = weight_s_scale(substitutions, 4)
        self.assertEqual(weight, Decimal('1.5'))
        weight = weight_s_scale(substitutions, 5, 2)
        self.assertEqual(weight, Decimal('1.5'))



#!/usr/bin/env python
'''
Setup script for installing the segeval package.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
import sys
import os

try:
    from setuptools import setup
    extra = dict(test_suite="segeval", include_package_data=True)
except ImportError:
    from distutils.core import setup
    extra = {}

from segeval import __package__, __version__, __author__, __author_email__, __description__

packages=['segeval',
          'segeval.agreement',
          'segeval.data',
          'segeval.ml',
          'segeval.similarity',
          'segeval.similarity.distance',
          'segeval.util',
          'segeval.window']

setup(
    name=__package__,
    version=__version__,
    long_description=open('README.rst').read(),
    description=__description__,
    license=open('LICENSE').read(),
    author=__author__,
    author_email=__author_email__,
    url='http://segeval.readthedocs.org/',
    # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=['Development Status :: 5 - Production/Stable',
                 'Environment :: Console',
                 'Intended Audience :: Developers',
                 'Intended Audience :: Science/Research',
                 'Natural Language :: English',
                 'License :: OSI Approved :: BSD License',
                 'Operating System :: OS Independent',
                 'Programming Language :: Python',
                 'Topic :: Scientific/Engineering :: Artificial Intelligence',
                 'Topic :: Scientific/Engineering :: Information Analysis',
                 'Topic :: Text Processing',
                 'Topic :: Utilities',
                 'Programming Language :: Python :: 2.6',
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3.2',
                 'Programming Language :: Python :: 3.3'],
      platforms = ['Any'],
      keywords = ['segmentation', 'similarity', 'discourse'],
      packages=packages,
      **extra)



'''
Multiple-boundary edit distance.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
from __future__ import division
from itertools import permutations
from collections import namedtuple


Addition = namedtuple('Addition', 'type side')  # For side; a = from a, b = from b
Substitution = namedtuple('Substitution', 'type_a type_b')
Transposition = namedtuple('Transposition', 'start end type')
Difference = namedtuple('Difference', 'sim a_b b_a')


def __additions_substitutions__(d, a, b):
    '''
    Compute the number of additions and substitutions for a given pair of
    boundary string positions using:

    :param d: Symmetric difference between the pair.
    :param a: Difference between the pair; A / B.
    :param b: Difference between the pair; B / A.
    :type d: set
    :type a: set
    :type b: set
    '''

    additions = abs(len(a) - len(b))
    substitutions = (len(d) - additions) / 2
    return additions, substitutions


def __additions_substitutions_sets__(d, a, b):
    '''
    Compute the sets of additions and substitutions for a given pair of
    boundary string positions using:

    :param d: Symmetric difference between the pair.
    :param a: Difference between the pair; A / B.
    :param b: Difference between the pair; B / A.
    :type d: set
    :type a: set
    :type b: set
    '''

    substitutions = list()
    delta = None
    for perm_a in permutations(sorted(a)):
        for perm_b in permutations(sorted(b)):
            current_substitutions = zip(perm_a, perm_b)
            current_substitutions = set(current_substitutions)
            current_delta = sum(abs(a_i - b_i)
                                for a_i, b_i in current_substitutions)
            if delta is None or current_delta < delta:
                delta = current_delta
                substitutions = current_substitutions
    # Collect all substitutions
    substituted = list()
    added = list()
    for a_i, b_i in substitutions:
        substituted.append(a_i)
        substituted.append(b_i)
    additions = d - set(substituted)
    # Add from a
    for addition in a - set(substituted):
        added.append(Addition(addition, 'a'))
    # Add from b
    for addition in b - set(substituted):
        added.append(Addition(addition, 'b'))
    assert len(added) is len(additions)
    return added, set([Substitution(a_i, b_i) for a_i, b_i in set(substitutions)])


def __has_substitutions__(i, j, d, options_set):
    '''
    Determine whether two substitutions are present involving the boundary 'd'
    at the positions 'i' and 'j'.
    '''

    present = False
    if i in options_set and d in options_set[i][0] and j in options_set and d in options_set[j][0]:
        d_i, a_i, b_i = options_set[i]
        d_j, a_j, b_j = options_set[j]
        if __additions_substitutions__(d_i, a_i, b_i)[1] > 0 and \
           __additions_substitutions__(d_j, a_j, b_j)[1] > 0:
            present = True
    return present


def __overlaps_existing__(i, j, d, options_transp):
    '''
    Determine whether a  transposition would overlap another that has been
    previously identified involving the boundary 'd' at the positions 'i' and
    'j'.
    '''

    def check_position(position):
        '''
        Check a position for an overlapping transposition
        '''
        if position in options_transp:
            for t in options_transp[position]:
                # If the transposition 't' is of the boundary type 'd'
                if t[2] is d:
                    return True
    return check_position(i) or check_position(j)


def __transpositions__(boundary_string_a, boundary_string_b, n, options_set):
    '''
    Identify all non-overlapping minimal transpositions between two boundary
    strings while removing the additions/deletions and substitutions that they
    would overlap from the set of potential additions/deletions and
    substitutions.
    '''

    options_transp = dict()
    transpositions = list()
    for n_i in sorted(n):
        n_i = n_i - 1
        for i in range(0, len(boundary_string_a) - n_i):
            j = i + n_i
            # Select edge sets
            a_i = boundary_string_a[i]
            a_j = boundary_string_a[j]
            b_i = boundary_string_b[i]
            b_j = boundary_string_b[j]
            # Compute symmetric differences
            diff_i = a_i ^ b_i
            diff_j = a_j ^ b_j
            diff_a = a_i ^ a_j
            diff_b = b_i ^ b_j
            # Detect potential transposition
            t_p = diff_i & diff_j & diff_a & diff_b
            # Apply each transposition found by boundary type
            for d in t_p:
                # Create transposition representation
                option_transp = Transposition(i, j, d)
                # Check to see that it does not overlap an existing
                # transposition and that 2 substitutions are not removed
                if not __overlaps_existing__(i, j, d, options_transp) and \
                   not __has_substitutions__(i, j, d, options_set):
                    # Add
                    transpositions.append(option_transp)
                    # Record positions covered
                    if i not in options_transp:
                        options_transp[i] = list()
                    if j not in options_transp:
                        options_transp[j] = list()
                    options_transp[i].append(option_transp)
                    options_transp[j].append(option_transp)
                    # Removing potential set errors that overlap
                    options_set[i][0].discard(d)
                    options_set[i][1].discard(d)
                    options_set[i][2].discard(d)
                    options_set[j][0].discard(d)
                    options_set[j][1].discard(d)
                    options_set[j][2].discard(d)
    return transpositions


def __optional_set_edits__(boundary_string_a, boundary_string_b):
    '''
    Identify all potential additions/deletions and substitutions.
    '''

    options_set = dict()
    for i, value in enumerate(zip(boundary_string_a, boundary_string_b)):
        a_i, b_i = value
        a = set(a_i - b_i)
        b = set(b_i - a_i)
        d = set(a_i ^ b_i)
        # Record additions/deletions
        if len(d) > 0:
            options_set[i] = Difference(d, a, b)
    return options_set


def __boundary_edit_distance__(boundary_string_a, boundary_string_b, n_t):
    '''
    Identify the minimum set of additions, substitutions, and transpositions
    that could be applied between two boundary strings for a given set
    of transpositions spanning lengths 'n_t'.

    :param n_t: transposition spanning sizes allowed
    :type n_t:  list or set
    '''

    # Find potential addition/deletion/substitution operations
    options_set = __optional_set_edits__(boundary_string_a,
                                         boundary_string_b)
    # Find transpositions
    transpositions = __transpositions__(boundary_string_a,
                                        boundary_string_b, n_t, options_set)
    # Construct additions and substitutions
    additions = list()
    substitutions = list()
    for option in options_set.values():
        current_additions, current_substitutions = \
            __additions_substitutions_sets__(option.sim, option.a_b, option.b_a)
        additions.extend(current_additions)
        substitutions.extend(current_substitutions)
    # Return
    return additions, substitutions, transpositions


def boundary_edit_distance(boundary_string_a, boundary_string_b, n_t=2):
    '''
    Computes boundary edit distance between two boundary strings.  Returns a \
    list of Addition, Substitution, and Transposition edit sets.

    :param boundary_string_a: Boundary string to compare; produced by :func:`boundary_string_from_masses`
    :param boundary_string_b: See `boundary_string_a`
    :param n_t: Maximum distance (in potential boundary positions) that a \
        transposition may span

    :type boundary_string_a:  tuple
    :type boundary_string_b:  tuple
    :type n_t:  int
    '''

    n_t = range(2, n_t + 1)
    return __boundary_edit_distance__(boundary_string_a, boundary_string_b, n_t)



'''
Arstein Poesio's annotator bias [ArtsteinPoesio2008]_.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
from . import __fnc_metric__
from .kappa import __fleiss_kappa_linear__
from .pi import __fleiss_pi_linear__


def __artstein_poesio_bias_linear__(dataset, **kwargs):
    '''
    Artstein and Poesio's annotator bias [ArtsteinPoesio2008]_.
    '''
    metric_kwargs = dict(kwargs)
    metric_kwargs['return_parts'] = True
    # Arguments
    return_parts = kwargs['return_parts']
    # Compute
    A_pi_e = __fleiss_pi_linear__(dataset, **metric_kwargs)[1]
    A_fleiss_e = __fleiss_kappa_linear__(dataset, **metric_kwargs)[1]
    bias = A_pi_e - A_fleiss_e
    # Return
    if return_parts:
        return A_pi_e, A_fleiss_e
    else:
        return bias


def artstein_poesio_bias_linear(dataset, **kwargs):
    '''
    Artstein and Poesio's annotator bias [ArtsteinPoesio2008]_.
    '''
    return __fnc_metric__(__artstein_poesio_bias_linear__, dataset, **kwargs)



'''
Tests multiple-boundary edit distance.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
import unittest
from .multipleboundary import (boundary_edit_distance,
                               __additions_substitutions__, __additions_substitutions_sets__,
                               __has_substitutions__)


class TestMultipleBoundaries(unittest.TestCase):

    '''
    Test multiple boundary edit distance.
    '''

    def test_edit_distance_identity(self):
        '''
        Test a mixed example of additions, substitutions and transpositions.
        '''
        a = [set(), set([1]), set(), set(), set([1]), set(), set(), set(), set(), set()]
        b = a

        additions, substitutions, transpositions = boundary_edit_distance(a, b)
        self.assertEqual(([], [], []),
                         (additions, substitutions, transpositions))

    def test_edit_distance_compliment(self):
        '''
        Test a mixed example of additions, substitutions and transpositions.
        '''
        a = [set(), set(), set(), set(), set(), set(), set(), set(), set(), set()]
        b = [set([1]), set([1]), set([1]), set([1]), set([1]), set([1]), set([1]), set([1]), set([1]), set([1])]

        additions, substitutions, transpositions = boundary_edit_distance(a, b)
        self.assertEqual(([(1, 'b'), (1, 'b'), (1, 'b'), (1, 'b'), (1, 'b'), (1, 'b'), (1, 'b'), (1, 'b'), (1, 'b'), (1, 'b')], [], []),
                         (additions, substitutions, transpositions))

    def test_edit_distance(self):
        '''
        Test a mixed example of additions, substitutions and transpositions.
        '''
        a = [set(), set([1]), set(), set(), set([1]), set(), set(), set(), set(), set()]
        b = [set(), set([2, 3]), set(), set(), set(), set([1]), set(), set(), set([3]), set()]

        additions, substitutions, transpositions = boundary_edit_distance(a, b)
        self.assertEqual(([(3, 'b'), (3, 'b')], [(1, 2)], [(4, 5, 1)]),
                         (additions, substitutions, transpositions))

    def test_edit_distance_two_transpositions(self):
        '''
        Test two transpositions
        '''
        a = [set(), set(), set(), set([2]), set(), set(), set([2]), set(), set(), set()]
        b = [set(), set(), set(), set(), set(), set([2]), set(), set(), set(), set()]
        additions, substitutions, transpositions = boundary_edit_distance(a, b)
        self.assertEqual(([(2, 'a')], [], [(5, 6, 2)]),
                         (additions, substitutions, transpositions))

    def test_edit_distance_no_two_transpositions(self):
        '''
        Test two transpositions
        '''
        a = [set(), set(), set([2]), set(), set(), set([2]), set([2]), set(), set(), set()]
        b = [set(), set(), set(), set(), set(), set([2]), set([2]), set(), set(), set()]
        additions, substitutions, transpositions = boundary_edit_distance(a, b)
        self.assertEqual(([(2, 'a')], [], []),
                         (additions, substitutions, transpositions))

    def test_edit_distance_two_transpositions_equal(self):
        '''
        Test two transpositions
        '''
        a = [set(), set(), set(), set(), set([2]), set(), set([2]), set(), set(), set()]
        b = [set(), set(), set(), set(), set(), set([2]), set(), set(), set(), set()]
        additions, substitutions, transpositions = boundary_edit_distance(a, b)
        self.assertEqual(([(2, 'a')], [], [(4, 5, 2)]),
                         (additions, substitutions, transpositions))

    def test_edit_distance_two_substitutions_into_two_transpositions(self):
        '''
        Test two transpositions
        '''
        a = [set(), set(), set(), set([2]), set(), set([1]), set(), set(), set(), set()]
        b = [set(), set(), set(), set([1]), set(), set([2]), set(), set(), set(), set()]
        additions, substitutions, transpositions = boundary_edit_distance(a, b, n_t=3)
        self.assertEqual(([], [(2, 1), (1, 2)], []),
                         (additions, substitutions, transpositions))

    def test_edit_distance_two_substitutions_into_no_transpositions(self):
        '''
        Test two transpositions
        '''
        a = [set(), set(), set(), set([2]), set(), set([3]), set(), set(), set(), set()]
        b = [set(), set(), set(), set([1]), set(), set([2]), set(), set(), set(), set()]
        additions, substitutions, transpositions = boundary_edit_distance(a, b, n_t=3)
        self.assertEqual(([], [(2,1), (3,2)], []),
                         (additions, substitutions, transpositions))

    def test_edit_distance_two_substitutions_one_ad_into_one_transpositions(self):
        '''
        Test two transpositions
        '''
        a = [set(), set(), set(), set([2,4]), set(), set([3]), set(), set(), set(), set()]
        b = [set(), set(), set(), set([1]), set(), set([2]), set(), set(), set(), set()]
        additions, substitutions, transpositions = boundary_edit_distance(a, b, n_t=3)
        self.assertEqual(([(4, 'a')], [(2, 1), (3, 2)], []),
                         (additions, substitutions, transpositions))

    def test_edit_distance_two_substitutions_one_ad_into_no_transpositions(self):
        '''
        Test two transpositions
        '''
        a = [set(), set(), set(), set([2]), set(), set([3]), set(), set(), set(), set()]
        b = [set(), set(), set(), set([1,4]), set(), set([2]), set(), set(), set(), set()]
        additions, substitutions, transpositions = boundary_edit_distance(a, b, n_t=3)
        self.assertEqual(([(4, 'b')], [(2,1), (3,2)], []),
                         (additions, substitutions, transpositions))

    def test_edit_distance_three_transpositions_overlapping(self):
        '''
        Test two transpositions
        '''
        a = [set(), set(), set(), set([1,2,3]), set(),    set(), set(), set(), set(), set()]
        b = [set(), set(), set(), set(), set([3]), set([1,2]), set(), set(), set(), set()]
        additions, substitutions, transpositions = boundary_edit_distance(a, b, n_t=3)
        self.assertEqual(([], [], [(3,4,3),(3,5,1),(3,5,2)]),
                         (additions, substitutions, transpositions))


class TestInnerFncOfMultipleBoundaries(unittest.TestCase):

    '''
    Test private functions for multiple boundary edit distance calculation.
    '''

    def test_has_substitutions_not_present(self):
        '''
        '''
        self.assertFalse(__has_substitutions__(1, 1, 1, set()))

    def test_additions_substitutions(self):
        '''
        Test the expected functionality of __additions_substitutions__
        '''
        a_i = set([1, 2, 3, 4])
        b_i = set([1, 6])

        a = a_i - b_i
        b = b_i - a_i
        d = a_i ^ b_i

        self.assertEqual((2, 1),
                         __additions_substitutions__(d, a, b))

    def test_additions_substitutions_sets(self):
        '''
        Test the expected functionality of __additions_substitutions_sets__
        '''
        a_i = set([1, 2, 3, 4])
        b_i = set([1, 6])

        a = a_i - b_i
        b = b_i - a_i
        d = a_i ^ b_i

        self.assertEqual(([(2, 'a'), (3, 'a')], set([(4, 6)])),
                         __additions_substitutions_sets__(d, a, b))



'''
Tests the WindowDiff evaluation metric.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
from decimal import Decimal
from ..compute import summarize
from ..format import BoundaryFormat
from ..window.pk import pk
from ..data.samples import (KAZANTSEVA2012_G5, KAZANTSEVA2012_G2,
                            COMPLETE_AGREEMENT, LARGE_DISAGREEMENT)
from ..util import SegmentationMetricError
from ..util.test import TestCase
from ..data.samples import HEARST_1997_STARGAZER, HYPOTHESIS_STARGAZER


class TestPk(TestCase):

    '''
    Test Pk.
    '''

    kwargs = {'boundary_format': BoundaryFormat.position}

    def test_one_minus(self):
        '''
        Test one minus.
        '''
        value = pk([2, 3, 6], [2, 2, 7], one_minus=True)
        self.assertAlmostEqual(Decimal('0.77777777'), value)

    def test_return_parts(self):
        '''
        Test one minus.
        '''
        value = pk(KAZANTSEVA2012_G5, return_parts=True)
        self.assertEqual((3, 10), value['ch1,an3,an1'])

    def test_return_parts_dataset(self):
        '''
        Test one minus.
        '''
        value = pk([2, 3, 6], [2, 2, 7], return_parts=True)
        self.assertEqual((2, 9), value)

    def test_identical(self):
        '''
        Test whether identical segmentations produce 0.0.
        '''

        a = [1,1,1,1,1,2,2,2,3,3,3,3,3]
        b = [1,1,1,1,1,2,2,2,3,3,3,3,3]
        one_minus_kwargs = dict(TestPk.kwargs)
        one_minus_kwargs['one_minus'] = True
        self.assertEqual(pk(a, b, **self.kwargs), Decimal('0.0'))
        self.assertEqual(pk(a, b, **one_minus_kwargs), Decimal('1.0'))

    def test_no_boundaries(self):
        '''
        Test whether no segments versus some segments produce 1.0.
        '''

        a = [1,1,1,1,1,1,1,1,1,1,1,1,1]
        b = [1,1,1,1,2,2,2,2,3,3,3,3,3]
        self.assertEqual(pk(b, a, **self.kwargs),
                         Decimal('1.0'))
        self.assertEqual(pk(a, b, **self.kwargs),
                         Decimal('0.3636363636363636363636363636'))

    def test_all_boundaries(self):
        '''
        Test whether all segments versus some segments produces 7/11 = 0.636
        erroneous windows.
        '''

        a = [1,2,3,4,5,6,7,8,9,10,11,12,13]
        b = [1,1,1,1,2,2,2,2,3,3,3,3,3]
        self.assertEqual(pk(a, b, **self.kwargs),
                         Decimal('0.6363636363636363636363636364'))
        self.assertEqual(pk(b, a, **self.kwargs),
                         Decimal('0.6363636363636363636363636364'))

    def test_all_and_no_boundaries(self):
        '''
        Test whether all segments versus no segments produces 1.0.
        '''

        a = [1,2,3,4,5,6,7,8,9,10,11,12,13]
        b = [1,1,1,1,1,1,1,1,1,1,1,1,1]
        self.assertEqual(pk(a, b, **self.kwargs), Decimal('1.0'))
        self.assertEqual(pk(b, a, **self.kwargs), Decimal('1.0'))

    def test_translated_boundary(self):
        '''
        Test whether 2/3 total segments participate in mis-alignment produces
        0.182.
        '''

        a = [1,1,1,1,1,2,2,2,3,3,3,3,3]
        b = [1,1,1,1,2,2,2,2,3,3,3,3,3]
        self.assertEqual(pk(a, b, **self.kwargs),
                         Decimal('0.1818181818181818181818181818'))
        self.assertEqual(pk(b, a, **self.kwargs),
                         Decimal('0.1818181818181818181818181818'))

    def test_extra_boundary(self):
        '''
        Test whether 1/3 segments that are non-existent produces 0.091.
        '''

        a = [1,1,1,1,1,2,2,2,3,3,3,3,3]
        b = [1,1,1,1,1,2,3,3,4,4,4,4,4]
        self.assertEqual(pk(a, b, **self.kwargs),
                         Decimal('0.09090909090909090909090909091'))
        self.assertEqual(pk(b, a, **self.kwargs),
                         Decimal('0.09090909090909090909090909091'))

    def test_full_miss_and_misaligned(self):
        '''
        Test whether a full miss and a translated boundary out of 4 produces
        0.273.
        '''

        a = [1,1,1,1,2,2,2,2,3,3,3,3,3]
        b = [1,1,1,1,1,2,3,3,4,4,4,4,4]
        self.assertEqual(pk(a, b, **self.kwargs),
                         Decimal('0.2727272727272727272727272727'))
        self.assertEqual(pk(b, a, **self.kwargs),
                         Decimal('0.2727272727272727272727272727'))

    def test_all_kwargs_hyp_ref(self):
        '''
        Test whether a full miss and a translated boundary out of 4 produces
        0.273.
        '''

        metric_kwargs = dict(self.kwargs)
        metric_kwargs['hypothesis'] = [1,1,1,1,2,2,2,2,3,3,3,3,3]
        metric_kwargs['reference'] = [1,1,1,1,1,2,3,3,4,4,4,4,4]
        self.assertEqual(pk(**metric_kwargs),
                         Decimal('0.2727272727272727272727272727'))

    def test_parts(self):
        '''
        Test parts.
        '''

        a = [1,2,3,4,5,6,7,8,9,10,11,12,13]
        b = [1,1,1,1,2,2,2,2,3,3,3,3,3]
        metric_kwargs = dict(self.kwargs)
        metric_kwargs['return_parts'] = True
        self.assertEqual(pk(a, b, **metric_kwargs),
                         (7, 11))

    def test_format_exception(self):
        '''
        Test format exception.
        '''
        a = [2, 3, 6]
        b = [2, 2, 7]
        self.assertRaises(SegmentationMetricError, pk,
                          a, b, boundary_format=BoundaryFormat.sets)

    def test_mass_exception(self):
        '''
        Test length mismatch exception.
        '''
        a = [2, 2, 7]
        b = [2, 2, 8]
        self.assertRaises(SegmentationMetricError, pk, a, b)

    def test_window_size_specified(self):
        '''
        Test when window size is specified.
        '''
        value = pk([2, 3, 6], [2, 2, 7], window_size=2)
        self.assertAlmostEqual(Decimal('0.2222222'), value)

    def test_boundary_format_nltk(self):
        '''
        Test the nltk boundary format.
        '''
        value = pk('0100100000', '0101000000', window_size=2, boundary_format=BoundaryFormat.nltk)
        self.assertAlmostEqual(Decimal('0.2222222'), value)

    def test_nltk(self):
        '''
        Runs Pk tests from https://github.com/nltk/nltk/blob/master/nltk/test/segmentation.doctest
        '''
        # Originally 0.0
        self.assertAlmostEqual(
            pk('1000100', '1000100', window_size=3, boundary_format=BoundaryFormat.nltk),
            Decimal('0.0'))
        # Originally 0.5
        self.assertAlmostEqual(
            pk('010', '100', window_size=2, boundary_format=BoundaryFormat.nltk),
            Decimal('0.5'))
        # Originally 0.64
        self.assertAlmostEqual(
            pk('111111', '100100', window_size=2, boundary_format=BoundaryFormat.nltk),
            Decimal('0.4'))
        # Originally 0.04
        self.assertAlmostEqual(
            pk('000000', '100100', window_size=2, boundary_format=BoundaryFormat.nltk),
            Decimal('0.6'))
        # Originally 0.25
        self.assertAlmostEqual(
            pk('111111', '100100', window_size=3, boundary_format=BoundaryFormat.nltk),
            Decimal('0'))
        # Originally 0.25
        self.assertAlmostEqual(
            pk('000000', '100100', window_size=3, boundary_format=BoundaryFormat.nltk),
            Decimal('1'))


class TestPairwisePkMeasure(TestCase):

    '''
    Test pairwise Pk.
    '''

    def test_kazantseva2012_g5(self):
        '''
        Calculate permuted pairwise Pk on Group 5 from the dataset
        collected in [KazantsevaSzpakowicz2012]_.
        '''
        self.assertAlmostEquals(summarize(pk(KAZANTSEVA2012_G5)),
                               (Decimal('0.35530058282396693'),
                                Decimal('0.11001760846099215'),
                                Decimal('0.012103874171476172'),
                                Decimal('0.015879673965138168'),
                                48))

    def test_kazantseva2012_g2(self):
        '''
        Calculate mean permuted pairwise Pk on Group 2 from the dataset
        collected in [KazantsevaSzpakowicz2012]_.
        '''
        self.assertAlmostEquals(summarize(pk(KAZANTSEVA2012_G2)),
                               (Decimal('0.2882256923776327507173609771'),
                                Decimal('0.1454395656787966169084191445'),
                                Decimal('0.02115266726483699483402909754'),
                                Decimal('0.01327675514600517730547602481'),
                                120))

    def test_large_disagreement(self):
        '''
        Calculate mean permuted pairwise Pk on a theoretical dataset
        containing large disagreement.
        '''
        self.assertAlmostEquals(summarize(pk(LARGE_DISAGREEMENT)),
                               (1.0,
                                0.0,
                                0.0,
                                0.0,
                                8))

    def test_complete_agreement(self):
        '''
        Calculate mean permuted pairwise Pk on a theoretical dataset
        containing complete agreement.
        '''
        self.assertAlmostEquals(summarize(pk(COMPLETE_AGREEMENT)),
                               (0.0,
                                0.0,
                                0.0,
                                0.0,
                                48))

    def test_dataset_kwargs(self):
        '''
        Calculate mean permuted pairwise Pk on a theoretical dataset
        containing complete agreement.
        '''
        self.assertAlmostEquals(summarize(pk(dataset=COMPLETE_AGREEMENT)),
                               (0.0,
                                0.0,
                                0.0,
                                0.0,
                                48))

    def test_pk_datasets(self):
        '''
        Test pk upon two datasets.
        '''
        hypothesis = HYPOTHESIS_STARGAZER
        reference = HEARST_1997_STARGAZER
        value = pk(hypothesis, reference)
        self.assertAlmostEquals(float(value['stargazer,h1,1']), 0.26315789)
        self.assertAlmostEquals(float(value['stargazer,h2,1']), 0.36842105)
        self.assertAlmostEquals(float(value['stargazer,h1,2']), 0.42105263)
        self.assertAlmostEquals(float(value['stargazer,h2,2']), 0.42105263)



'''
Inter-coder agreement statistic Fleiss' Kappa.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
from __future__ import division
from decimal import Decimal
from . import __fnc_metric__, __actual_agreement_linear__


def __fleiss_kappa_linear__(dataset, **kwargs):
    '''
    Calculates Fleiss' :math:`\kappa` (or multi-:math:`\kappa`), originally proposed in
    [DaviesFleiss1982]_.  For 2 coders, this is equivalent to Cohen's :math:`\kappa`
    [Cohen1960]_.
    '''
    metric_kwargs = dict(kwargs)
    metric_kwargs['return_parts'] = True
    # Arguments
    return_parts = kwargs['return_parts']
    # Check that there are more than 2 coders
    if len([True for coder_segs in dataset.values()
            if len(coder_segs.keys()) < 2]) > 0:
        raise Exception('Less than 2 coders specified.')
    # Check that there are an identical number of items
    if len(set([len(coder_segs.values()) for coder_segs in dataset.values()])) != 1:
        raise Exception('Unequal number of items contained.')
    # Initialize totals
    all_numerators, all_denominators, _, coders_boundaries = \
        __actual_agreement_linear__(dataset, **metric_kwargs)
    # Calculate Aa
    A_a = Decimal(sum(all_numerators)) / sum(all_denominators)
    # Calculate Ae
    coders = list(coders_boundaries.keys())
    P_segs = list()
    for m in range(0, len(coders) - 1):
        for n in range(m + 1, len(coders)):
            boundaries_m = sum(info[0] for info in
                               coders_boundaries[coders[m]])
            total_boundaries_m = sum(info[1] for info in
                                     coders_boundaries[coders[m]])
            boundaries_n = sum(info[0] for info in
                               coders_boundaries[coders[n]])
            total_boundaries_n = sum(info[1] for info in
                                     coders_boundaries[coders[n]])
            P_segs.append((Decimal(boundaries_m) / total_boundaries_m) *
                          (Decimal(boundaries_n) / total_boundaries_n))
    P_seg = Decimal(sum(P_segs)) / len(P_segs)
    A_e = P_seg
    # Calculate pi
    kappa = (A_a - A_e) / (Decimal('1') - A_e)
    # Return
    if return_parts:
        return A_a, A_e
    else:
        return kappa


def fleiss_kappa_linear(dataset, **kwargs):
    '''
    Calculates Fleiss' :math:`\kappa` (or multi-:math:`\kappa`), originally proposed in
    [DaviesFleiss1982]_.  For 2 coders, this is equivalent to Cohen's :math:`\kappa`
    [Cohen1960]_.
    '''
    return __fnc_metric__(__fleiss_kappa_linear__, dataset, **kwargs)
